#include <math.h>
#include <stdio.h>
int matinv(double *a, double *b, int ijob, int n){
    /*   *DECK MATINV
C
C   PURPOSE             - IN PLACE INVERSE, EQUATION SOLUTION, POSITIVE
C                           DEFINITE MATRIX - SYMMETRIC STORAGE MODE
C
C   USAGE               - CALL MATINV (A,B,IJOB,N,IER)  c: ier=matinv(x,y,ijob);
C
C   ARGUMENTS    A      - INPUT/OUTPUT VECTOR OF LENGTH N(N+1)/2.
C                         ON INPUT, A CONTAINS THE N BY N POSITIVE
C                           DEFINITE SYMMETRIC MATRIX STORED IN
C                           SYMMETRIC STORAGE MODE.
C                         ON OUTPUT, SEE PARAMETER IJOB.
C                B      - INPUT/OUTPUT VECTOR OF LENGTH N WHEN IJOB =
C                           2 OR 3.
C                         ON INPUT, B CONTAINS THE RIGHT HAND SIDE OF
C                           THE EQUATION AX=B.
C                         ON OUTPUT, THE SOLUTION X REPLACES B WHEN
C                           IJOB=2 OR 3. OTHERWISE, B IS NOT USED.
C                IJOB   - INPUT OPTION PARAMETER. IJOB = I IMPLIES:
C                           I = 1, INVERT MATRIX A. A IS REPLACED BY ITS
C                             INVERSE STORED IN SYMMETRIC STORAGE MODE.
C                           I = 2, SOLVE THE EQUATION AX = B. A IS
C                             REPLACED BY THE DECOMPOSED MATRIX L SUCH
C                             THAT A = L*L-TRANSPOSE. L IS STORED IN
C                             SYMMETRIC STORAGE MODE. THE DIAGONAL OF L
C                             CONTAINS THE RECIPROCALS OF THE ACTUAL
C                             DIAGONAL ELEMENTS OF L.
C                           I = 3, SOLVE AX = B AND INVERT MATRIX A.
C                             A IS REPLACED BY ITS INVERSE AND THE
C                             SOLUTION X REPLACES B.
C                N      - ORDER OF A. (INPUT)
C                IER    - ERROR PARAMETER. (OUTPUT)
C                         TERMINAL ERROR
C                           IER = 129 INDICATES THAT IJOB WAS LESS THAN
C                             1 OR GREATER THAN 3.
C                           IER = 130 INDICATES THAT MATRIX A IS
C                             ALGORITHMICALLY NOT POSITIVE DEFINITE.
C                             (SEE THE CHAPTER L PRELUDE).
C-----------------------------------------------------------------------
      SUBROUTINE MATINV (A,B,IJOB,N,IER)
*/
	void srverror(char *x); void srvwarn(char *x);
    double q,rn,s,t,x; 
    int ier=0,isw,i,j,k,l,ii,ip,is,iq,kk,ip1,nm1,jm1,jl,li,iw,im1,n1,lj,ji,ir,jj; char tmpbuf[128];
    
    if (ijob<1 || ijob>3) { 
		ier=129; sprintf(tmpbuf,"*** ERROR MESSAGE IER=%d FROM ROUTINE matinv\n",ier); 
		srvwarn(tmpbuf); return(ier); 
	}
    isw=0; rn=1./n/16.; ip=1; ier=0;
    for (i=1; i<=n; i++) {
        iq=ip; ir=1;
        for (j=1; j<=i; j++,ir++) {
			x=a[ip];
            if (j!=1) for (k=iq; k<=ip1; k++) x-=a[k]*a[ir++];
            if (i==j) { 
                q=a[ip]+x*rn; 
                if (q<=a[ip]) {
					ier=130; sprintf(tmpbuf,"*** ERROR MESSAGE IER=%d FROM ROUTINE matinv\n",ier); 
					srvwarn(tmpbuf); return(ier); 
				}
                a[ip]=1/sqrt(x);
            }
            else a[ip]=x*a[ir];
			ip1=ip++; 
        }
    }
    if (ijob!=1 && isw!=1) {       // SOLVE AX = B
        isw=1; ip=1; iw=0; im1=0;     //   SOLUTION OF LY = B
        for (i=1; i<=n; i++) {
            t=b[i]; 
            if (iw!=0) {
                ip+=(iw-1);
                for (k=iw; k<=im1; k++) t-=a[ip++]*b[k];
            }
            else {
                if (t!=0) iw=i;
                ip+=im1;
            }
            b[i]=t*a[ip++]; im1=i;
        }
        n1=n+1;    // SOLUTION OF UX = Y WHERE U = L-TRANSPOSE
        for (i=1; i<=n; i++) {
            ii=n1-i; is=--ip; iq=ii+1; t=b[ii];
            if (n>=iq) for (k=iq,kk=n; k<=n; k++) { t-=a[is]*b[kk--]; is-=kk; }
            b[ii]=t*a[is];
        }
    }
    if (isw!=0 || ijob==1 || ijob==3) {
        nm1=n-1;         // FORM A-INVERSE
        if (n!=1) {
            ii=1;              // COMPUTE L-INVERSE FIRST
            for (i=1; i<=nm1; i++) {
                ip1=i+1; jm1=i; jj=ii;
                for (j=ip1; j<=n; j++) {
                    s=0; li=ii; ji=jj+i; jl=ji;
                    for (l=i; l<=jm1; l++) { s+=a[li]*a[jl++]; li+=l;}
                    jj+=j; a[ji]=-a[jj]*s; jm1=j;
                }
                ii+=ip1;
            }
        }
        ii=0;                  // NOW FORM A-INVERSE
        for (i=1; i<=n; i++) {
            for (j=i,jj=ii; j<=n; j++) {
                s=0; ji=jj+i; li=ji; jj+=j; lj=jj;
                for (l=j; l<=n; l++) { s+=a[li]*a[lj]; li+=l; lj+=l; }
                a[ji]=s;
            }
			ii+=i;
        }                              
    }
	return(ier);
}
