#include <math.h>
#include <stdio.h>
double dabs (double x) ;
double dmax (double x, double y) ;
double est(int nest, double *ppar);
void optmz2(double *a, int n, double *z, double sig, double *w, int *ir, int mk, double eps);
int optmiz(int n, int nsig, int maxfn, int iopt, double *x, double *h, double *g, double *rf, double *w) {
    /*      SUBROUTINE OPTMIZ MAXIMIZES THE LOG LIKELIHOOD FUNCTION
C     SUBROUTINE OPTMZ2 IS CALLED.
C     FUNCTION            - A QUASI-NEWTON ALGORITHM FOR FINDING THE
C                           MINIMUM OF A FUNCTION OF N VARIABLES.
C     USAGE               - CALL OPTMIZ(FUNCT,N,NSIG,MAXFN,IOPT,X,H,G,F,W,IER)
C     PARAMETERS   FUNCT  - A USER SUPPLIED SUBROUTINE WHICH CALCULATES
C                           THE FUNCTION F FOR GIVEN PARAMETER VALUES
C                           X(1),X(2),...,X(N).
C                           THE CALLING SEQUENCE HAS THE FOLLOWING FORM
C                           CALL FUNCT(N,X,F)
C                           WHERE X IS A VECTOR OF LENGTH N.
C                           FUNCT MUST APPEAR IN AN EXTERNAL STATEMENT
C                           IN THE CALLING PROGRAM. FUNCT MUST NOT
C                           ALTER THE VALUES OF X(I),I=1,...,N OR N.
C                N      - THE NUMBER OF PARAMETERS (I.E., THE LENGTH
C                           OF X) (INPUT)
C                NSIG   - CONVERGENCE CRITERION. (INPUT). THE NUMBER
C                           OF DIGITS OF ACCURACY REQUIRED IN THE
C                           PARAMETER ESTIMATES.
C                           THIS CONVERGENCE CONDITION IS SATISIFIED IF
C                           ON TWO SUCCESSIVE ITERATIONS, THE PARAMETER
C                           ESTIMATES (I.E.,X(I), I=1,...,N) AGREE,
C                           COMPONENT BY COMPONENT, TO NSIG DIGITS.
C                MAXFN  - MAXIMUM NUMBER OF FUNCTION EVALUATIONS (I.E.,
C                           CALLS TO SUBROUTINE FUNCT) ALLOWED. (INPUT)
C                IOPT   - INPUT OPTIONS SELECTOR.
C                         IOPT = 0 CAUSES OPTMIZ TO INITIALIZE THE
C                           HESSIAN MATRIX H TO THE IDENTITY MATRIX.
C                         IOPT = 1 INDICATES THAT H HAS BEEN INITIALIZED
C                           BY THE USER TO A POSITIVE DEFINITE MATRIX.
C                X      - VECTOR OF LENGTH N CONTAINING PARAMETER
C                           VALUES.
C                         ON INPUT, X MUST CONTAIN THE INITIAL
C                           PARAMETER ESTIMATES.
C                         ON OUTPUT, X CONTAINS THE FINAL PARAMETER
C                           ESTIMATES AS DETERMINED BY OPTMIZ.
C                H      - VECTOR OF LENGTH N*(N+1)/2 CONTAINING AN
C                           ESTIMATE OF THE HESSIAN MATRIX
C                           D**2F/(DX(I)DX(J)), I,J=1,...,N.
C                           H IS STORED IN SYMMETRIC STORAGE MODE.
C                         ON INPUT, IF IOPT = 0, OPTMIZ INITIALIZES H
C                           TO THE IDENTITY MATRIX. AN INITIAL SETTING
C                           OF H BY THE USER IS INDICATED BY IOPT=1.
C                           H MUST BE POSITIVE DEFINITE. IF IT IS NOT,
C                           A TERMINAL ERROR OCCURS.
C                         ON OUTPUT, H CONTAINS AN ESTIMATE OF THE
C                           HESSIAN AT THE FINAL PARAMETER ESTIMATES
C                           (I.E., AT X(1),X(2),...,X(N))
C                G      - A VECTOR OF LENGTH N CONTAINING AN ESTIMATE
C                           OF THE GRADIENT DF/DX(I),I=1,...,N AT THE
C                           FINAL PARAMETER ESTIMATES. (OUTPUT)
C                F      - A SCALAR CONTAINING THE VALUE OF THE FUNCTION
C                           AT THE FINAL PARAMETER ESTIMATES. (OUTPUT)
C                W      - A VECTOR OF LENGTH 3*N USED AS WORKING SPACE.
C                         ON OUTPUT, WORK(I), CONTAINS FOR
C                           I = 1, THE NORM OF THE GRADIENT (I.E.,
C                             SQRT(G(1)**2+G(2)**2+...+G(N)**2))
C                           I = 2, THE NUMBER OF FUNCTION EVALUATIONS
C                             PERFORMED.
C                           I = 3, AN ESTIMATE OF THE NUMBER OF
C                             SIGNIFICANT DIGITS IN THE FINAL
C                             PARAMETER ESTIMATES.
C                IER    - ERROR PARAMETER.
C                         IER = 0 IMPLIES THAT CONVERGENCE WAS
C                           ACHIEVED AND NO ERRORS OCCURRED.
C                         TERMINAL ERROR
C                           IER = 129 IMPLIES THAT THE INITIAL HESSIAN
C                             MATRIX IS NOT POSITIVE DEFINITE. THIS
C                             CAN OCCUR ONLY FOR IOPT = 1.
C                           IER = 130 IMPLIES THAT THE ITERATION WAS
C                             TERMINATED DUE TO ROUNDING ERRORS
C                             BECOMING DOMINANT. THE PARAMETER
C                             ESTIMATES HAVE NOT BEEN DETERMINED TO
C                             NSIG DIGITS.
C                           IER = 131 IMPLIES THAT THE ITERATION WAS
C                             TERMINATED BECAUSE MAXFN WAS EXCEEDED.
C     PRECISION           - SINGLE
C     REQD.  ROUTINES     - OPTMZ2
C*/
    void srverror(char *x); void srvwarn(char *x);
    double hh,eps,hjj,v,df,relx,gs0,diff,aeps,alpha,ff,tot,f,f1,f2,z,gys,dgs,sig,zz,gnrm;
    int i,j,ij,ir,nm1,np1,jj,l,kj,itn,ifn,link,ig,jp1,jb,idiff,igg,is,k,ii,im1,nj,ier=0,jnt;
    double reps=1.45519152284e-11, ax=0.1, p1=0.1, tmpow=0;
	char tmpstr[128];
    
    ier=0; hh=sqrt(reps); eps=pow(10,-nsig); ig=n; igg=n+n; is=igg; idiff=1; ir=n;
    w[1]=-1; w[2]=0; w[3]=0;
    if (iopt==1) goto x30;
    for (ij=0, i=1; i<=n; i++) {  // SET H TO THE IDENTITY MATRIX
        for (j=1; j<=i; j++) h[++ij]=0;
        h[ij]=1;
    }
    goto x100;
x30:
    ir=n;         //  FACTOR H TO L*D*L-TRANSPOSE
    if (n>1) goto x40;
    if (h[1]>0) goto x100; 
    h[1]=0; ir=0; 
    goto x90;
x40:
    nm1=n-1; 
    for (jj=0, j=1; j<=n; j++) {
        jj+=j; hjj=h[jj];
        if (hjj<=0) { h[jj]=0; ir--;}
		else {
			if (j!=n) {
				for (ij=jj, l=0, i=j+1; i<=n; i++) {
					l++; ij+=i-1; v=h[ij]/hjj;
					for (kj=ij, k=i; k<=n; k++) { h[kj+l]+=h[kj]*v; kj+=k;}
					h[ij]=v;
				}
			}
		}
    }
x90:
    if (ir==n) goto x100;
    ier=129; goto x500;
x100:
    itn=0;  f=est(n,x);    //    EVALUATE FUNCTION AT STARTING POINT
    ifn=1; df=-1; 
x110:    
    link=1;    //  EVALUATE GRADIENT W(IG+I),I=1,...,N
    goto x460;
x120:       // BEGIN ITERATION LOOP
    if (ifn>=maxfn) goto x390;
    itn++;
    for (i=1; i<=n; i++) w[i]=-w[ig+i];
    if (ir<n) goto x190;
       //  DETERMINE SEARCH DIRECTION W BY SOLVING H*W = -G WHERE H = L*D*L-TRANSPOSE
    g[1]=w[1];
    if (n>1) goto x140;
    w[1]/=h[1];
    goto x190;
x140:
    for (ii=1, i=2; i<=n; i++) {  //  SOLVE L*W = -G
        ij=ii; ii+=i; v=w[i]; im1=i-1;
        for (j=1; j<=im1; j++) v-=h[++ij]*w[j];
        g[i]=v; w[i]=v;
    }
    w[n]/=h[ii];  // SOLVE (D*LT)*Z = W WHERE LT = L-TRANSPOSE
    jj=ii; nm1=n-1;
    for (nj=1; nj<=nm1; nj++) {
        j=n-nj; jp1=j+1; jj-=jp1; 
		v=w[j]/h[jj]; ij=jj;
        for (i=jp1; i<=n; i++) { ij+=i-1; v-=h[ij]*w[i]; }
        w[j]=v;
    }
x190:
    relx=0; gs0=0;
    for (i=1; i<=n; i++) {
        w[is+i]=w[i]; 
		diff=dabs(w[i])/dmax(dabs(x[i]),ax); 
        relx=dmax(relx,diff); gs0+=w[ig+i]*w[i];
    }
    if (relx==0) goto x400;
    aeps=eps/relx; ier=130;
    if (gs0>=0 || df==0) goto x400;
    ier=0; alpha=(-df-df)/gs0;
    if (alpha<=0 || alpha>1) alpha=1;
    if (idiff==2) alpha=dmax(p1,alpha);
    ff=f; tot=0; jnt=0;
x210:
    if (ifn>=maxfn) goto x390;
    for (i=1; i<=n; i++) w[i]=x[i]+alpha*w[is+i];
    f1=est(n,w); ifn++;
    if (f1>=f) goto x270;
    f2=f; tot+=alpha;
x230:
    ier=0; f=f1; for (i=1; i<=n; i++) x[i]=w[i];
    if ((jnt-1)==0) goto x310;
    if ((jnt-1)>0) goto x320;
    if (ifn>=maxfn) goto x390;
    for (i=1; i<=n; i++) w[i]=x[i]+alpha*w[is+i];
    f1=est(n,w);
    ifn++;
    if (f1>=f) goto x320;
    if ((f1+f2)>=(f+f) && (7*f1+5*f2)>(12*f)) jnt=2;
    tot+=alpha; alpha+=alpha;
    goto x230;
x270:
    if (f==ff && idiff==2 && relx>eps) ier=130;
    if (alpha<aeps) goto x400;
    if (ifn>=maxfn) goto x390;
    alpha=0.5*alpha;
    for (i=1; i<=n; i++) w[i]=x[i]+alpha*w[is+i];
    f2=est(n,w); ifn++;
    if (f2>=f) goto x300;
    tot+=alpha; ier=0; f=f2;
    for (i=1; i<=n; i++) x[i]=w[i];
    goto x310;
x300:
    z=p1; if ((f1+f)>(f2+f2)) z=1+0.5*(f-f1)/(f+f1-f2-f2);
    z=dmax(p1,z); alpha*=z; jnt=1;
    goto x210;
x310: 
    if (tot<aeps) goto x400;
x320:
    alpha=tot;
    for (i=1; i<=n; i++) w[i]=w[ig+i]; //  SAVE OLD GRADIENT
    link=2;    // EVALUATE GRADIENT W(IG+I), I=1,...,N
    goto x460;
x340:
    if (ifn>=maxfn) goto x390;
    gys=0;
    for (i=1; i<=n; i++) { gys+=w[ig+i]*w[is+i]; w[igg+i]=w[i]; }
    df=ff-f; dgs=gys-gs0;
    if (dgs<=0) goto x120;
    if ((dgs+alpha*gs0)>0) goto x370;
    sig=1/gs0; //   UPDATE HESSIAN H USING COMPLEMENTARY DFP FORMULA    
    ir=-ir;
    optmz2(h,n,w,sig,g,&ir,0,0);
    for (i=1; i<=n; i++) g[i]=w[ig+i]-w[igg+i];
    sig=1/(alpha*dgs);
    ir=-ir;
    optmz2(h,n,g,sig,w,&ir,0,0);
    goto x120;
x370:                     //   UPDATE HESSIAN USING DFP FORMULA
    zz=alpha/(dgs-alpha*gs0);
    sig=-zz;
    optmz2(h,n,w,sig,g,&ir,0,reps);
    z=dgs*zz-1;
    for (i=1; i<=n; i++) g[i]=w[ig+i]+z*w[igg+i];
    sig=1/(zz*dgs*dgs);
    optmz2(h,n,g,sig,w,&ir,0,0);
    goto x120;
x390:
    ier=131; 
    goto x410;
x400:
    if (idiff==2) goto x410;
    //                CHANGE TO CENTRAL DIFFERENCES
    idiff=2; goto x110;
x410:
    if (relx>eps && ier==0) goto x110;
    //             MOVE GRADIENT TO G AND RETURN
    for (gnrm=0, i=1; i<=n; i++) { g[i]=w[ig+i]; gnrm+=g[i]*g[i]; }
    w[1]=sqrt(gnrm); w[2]=ifn; w[3]=-log10(dmax(reps,relx));
    //             COMPUTE H = L*D*L-TRANSPOSE
    if (n==1) goto x500;
    np1=n+1; nm1=n-1;
    jj=n*np1/2;
    for (jb=1; jb<=nm1; jb++) {
        jp1=np1-jb; jj-=jp1; hjj=h[jj]; ij=jj; l=0;
        for (i=jp1; i<=n; i++) {
            l++; ij+=(i-1); v=h[ij]*hjj; kj=ij;
            for (k=i; k<=n; k++) { h[kj+l]+=h[kj]*v; kj+=k; }
            h[ij]=v;
        }
        hjj=h[jj];
    }
    goto x500;
x460:
    if (idiff==2) goto x480;
    for (i=1; i<=n; i++) {
        z=hh*dmax(dabs(x[i]),ax);
        zz=x[i]; x[i]=zz+z;
        f1=est(n,x);
        w[ig+i]=(f1-f)/z;
        x[i]=zz; 
    }
    ifn+=n;
    if (link==2) goto x340;
    goto x120;
x480:
    for (i=1; i<=n; i++) {
        z=hh*dmax(dabs(x[i]),ax);
        zz=x[i]; x[i]=zz+z;
        f1=est(n,x);
        x[i]=zz-z;
        f2=est(n,x);
        w[ig+i]=(f1-f2)/(z+z);
        x[i]=zz;
    }
    ifn+=n+n;
    if (link==2) goto x340;
    goto x120;
x500:
    if (ier!=0) {
		sprintf(tmpstr,"*** ERROR MESSAGE IER=%d FROM ROUTINE %s\n",ier,"OPTMIZ");
		srvwarn(tmpstr);
	}
    *rf=f; return(ier);
}        

